#include "../include/apsrstack-general.h"

void process_icmpv4_packet (struct apsrstack_frame_header *header, void *packet, int opt_size) {
  struct ipv4_pkt_header *ipv4_header;
  struct icmpv4_pkt_header *icmpv4_header;
  unsigned char *icmpv4_buffer;
  unsigned short payload_length = 0, old_sum = 0, calc_sum = 0;
  unsigned int receive_timestamp = 0, transmit_timestamp = 0;
  struct timeval tv;
      	
  gettimeofday(&(tv), NULL);
  receive_timestamp = apsr_htonl((tv.tv_sec % 86400) * 1000 + tv.tv_usec / 1000);
  transmit_timestamp = receive_timestamp; /* Not so good */
  
  ipv4_header = (struct ipv4_pkt_header *)(packet+data_link_size);
  icmpv4_header = (struct icmpv4_pkt_header *)(packet+data_link_size+opt_size);

  payload_length = apsr_ntohs(ipv4_header->packet_len) - (APSR_IPV4_HEADERLENGTH(ipv4_header)*4);
  
  /*
   * Checksum verification.
   */
  old_sum = icmpv4_header->sum;
  icmpv4_header->sum=0x0000;
  calc_sum=in_cksum((unsigned short *)icmpv4_header, payload_length, 0);
  if(old_sum != calc_sum) {
	  #ifdef APSRSTACK_DEBUG
	  apsr_output("DEBUG: Got ICMPv4 packet with wrong checksum.\n");
	  #endif
	  stats.icmpv4.checksum_errors++; /* Checksum Errors++*/
          return;
  }

  if(icmpv4_header->type == APSR_ICMP_ECHO_REQUE) {
	  if(icmpv4_header->code == 0) {
		  /* + 4 bytes for Identifier and Sequence Number */
		  payload_length = payload_length - APSR_ICMPv4 - 4;
	          icmpv4_buffer = (unsigned char *)(packet+data_link_size+opt_size+APSR_ICMPv4+4);
		  
		  send_icmpv4_echo(header,
				   5,
				   APSR_ICMPv4+4+payload_length,
				   apsr_ntohs(ipv4_header->ident),
				   ipv4_header->dst_ip,
				   ipv4_header->src_ip,
				   APSRSTACK_IPV4_DEFAULT_TTL,
				   APSR_ICMP_ECHO_REPLY,
				   0, 
				   apsr_ntohs(icmpv4_header->ext.echo.identifier),
				   apsr_ntohs(icmpv4_header->ext.echo.seq_num),
				   icmpv4_buffer,
				   payload_length);

		  stats.icmpv4.processed++; /* Counter */
		  return;
	  }
	  else { /* Unknown ICMPv4 Code */
	          #ifdef APSRSTACK_DEBUG
                  apsr_output("DEBUG: Unknown ICMPv4 code %u for type %u.\n", 
				  icmpv4_header->code, icmpv4_header->type);
	          #endif
  	   	  stats.icmpv4.drops++; /* Drops */
	          return;
	  }
  }
  else if(icmpv4_header->type == APSR_ICMP_TIME_REQUE) {	  
	  if(icmpv4_header->code == 0) {
		  send_icmpv4_timestamp(header,
				   5,
				   APSR_ICMPv4+16,
				   apsr_ntohs(ipv4_header->ident),
				   ipv4_header->dst_ip,
				   ipv4_header->src_ip,
				   APSRSTACK_IPV4_DEFAULT_TTL,
				   APSR_ICMP_TIME_REPLY,
				   0, 
				   apsr_ntohs(icmpv4_header->ext.timestamp.identifier),
				   apsr_ntohs(icmpv4_header->ext.timestamp.seq_num),
				   apsr_ntohl(icmpv4_header->ext.timestamp.originate),
				   receive_timestamp,
				   transmit_timestamp);

		  stats.icmpv4.processed++; /* Counter */
		  return;
	  }
	  else { /* Unknown ICMPv4 Code */
	          #ifdef APSRSTACK_DEBUG
                  apsr_output("DEBUG: Unknown ICMPv4 code %u for type %u.\n", 
				  icmpv4_header->code, icmpv4_header->type);
	          #endif
  	   	  stats.icmpv4.drops++; /* Drops */
	          return;
	  }
  }
  else { /* Unknown ICMPv4 Type */
	  #ifdef APSRSTACK_DEBUG
	  apsr_output("DEBUG: Unknown ICMPv4 type %u.\n",
			  icmpv4_header->type);
	  #endif			    
	  stats.icmpv4.drops++; /* Drops */
	  return;
  }
}

void send_icmpv4_timestamp(struct apsrstack_frame_header *header,
		 unsigned char ip_hdr_length,
		 unsigned char ip_length, unsigned short ip_ident,
		 struct in_addr ip_src, struct in_addr ip_dst, 
		 unsigned char ip_ttl, unsigned char icmp_type,
		 unsigned char icmp_code, unsigned short icmp_identifier,
		 unsigned short icmp_seq_num, unsigned int icmp_originate,
		 unsigned int icmp_receive, unsigned int icmp_transmit) {
	
	struct apsr_link_layer *apsr_ll;
	unsigned char *buf;

   	apsr_ll = (struct apsr_link_layer *) apsr_open_link_layer(device);
 
	apsr_init_psize(APSR_ETH + (ip_hdr_length * 4) + APSR_ICMPv4 + 16, &buf);

        apsr_ethernet(header->dst_mac, header->src_mac, header->type, NULL, 0, buf);
	
	apsr_ipv4(4, ip_hdr_length, ip_length, 0x0000, ip_ident, 0x0000, ip_ttl, APSR_PROTO_ICMP, ip_src.s_addr, ip_dst.s_addr, 0, NULL, 0, NULL, 0, buf + APSR_ETH);

	apsr_icmpv4_timestamp(icmp_type, icmp_code, 0, icmp_identifier, icmp_seq_num, icmp_originate, icmp_receive, icmp_transmit, NULL, 0, buf + APSR_ETH + (ip_hdr_length * 4));
	
        apsr_checksum(buf + APSR_ETH, APSR_PROTO_ICMP, APSR_ICMPv4 + 16);
	
	apsr_write_link_layer(apsr_ll, device, buf, APSR_ETH + (ip_hdr_length * 4) + APSR_ICMPv4 + 16, 1);
  
  	apsr_close_link_layer(apsr_ll);
        free(buf);
}

void send_icmpv4_echo(struct apsrstack_frame_header *header,
		 unsigned char ip_hdr_length,
		 unsigned char ip_length, unsigned short ip_ident,
		 struct in_addr ip_src, struct in_addr ip_dst, 
		 unsigned char ip_ttl, unsigned char icmp_type,
		 unsigned char icmp_code, unsigned short icmp_identifier,
		 unsigned short icmp_seq_num, unsigned char *icmp_payload,
		 unsigned short icmp_payload_length) {
	
	struct apsr_link_layer *apsr_ll;
	unsigned char *buf;

   	apsr_ll = (struct apsr_link_layer *) apsr_open_link_layer(device);
 
	apsr_init_psize(APSR_ETH + (ip_hdr_length * 4) + APSR_ICMPv4 + 4 + icmp_payload_length, &buf);

        apsr_ethernet(header->dst_mac, header->src_mac, header->type, NULL, 0, buf);
	
	apsr_ipv4(4, ip_hdr_length, ip_length, 0x0000, ip_ident, 0x0000, ip_ttl, APSR_PROTO_ICMP, ip_src.s_addr, ip_dst.s_addr, 0, NULL, 0, NULL, 0, buf + APSR_ETH);

	apsr_icmpv4_echo(icmp_type, icmp_code, 0, icmp_identifier, icmp_seq_num, icmp_payload, icmp_payload_length, buf + APSR_ETH + (ip_hdr_length * 4));
	
        apsr_checksum(buf + APSR_ETH, APSR_PROTO_ICMP, APSR_ICMPv4 + 4 + icmp_payload_length);
	
	apsr_write_link_layer(apsr_ll, device, buf, APSR_ETH + (ip_hdr_length * 4) + APSR_ICMPv4 + 4 + icmp_payload_length, 1);
  
  	apsr_close_link_layer(apsr_ll);
        free(buf);
}

void send_icmpv4(struct apsrstack_frame_header *header,
		 unsigned char ip_hdr_length,
		 unsigned char ip_length, unsigned short ip_ident,
		 struct in_addr ip_src, struct in_addr ip_dst, 
		 unsigned char ip_ttl, unsigned char icmp_type,
		 unsigned char icmp_code, unsigned char *icmp_payload,
                 unsigned short icmp_payload_length) {
	
	struct apsr_link_layer *apsr_ll;
	unsigned char *buf;

   	apsr_ll = (struct apsr_link_layer *) apsr_open_link_layer(device);
 
	apsr_init_psize(APSR_ETH + (ip_hdr_length * 4) + APSR_ICMPv4 + icmp_payload_length, &buf);

        apsr_ethernet(header->dst_mac, header->src_mac, header->type, NULL, 0, buf);
	
	apsr_ipv4(4, ip_hdr_length, ip_length, 0x0000, ip_ident, 0x0000, ip_ttl, APSR_PROTO_ICMP, ip_src.s_addr, ip_dst.s_addr, 0, NULL, 0, NULL, 0, buf + APSR_ETH);

	apsr_icmpv4(icmp_type, icmp_code, 0, icmp_payload, icmp_payload_length, buf + APSR_ETH + (ip_hdr_length * 4));
	
        apsr_checksum(buf + APSR_ETH, APSR_PROTO_ICMP, APSR_ICMPv4 + icmp_payload_length);
	
	apsr_write_link_layer(apsr_ll, device, buf, APSR_ETH + (ip_hdr_length * 4) + APSR_ICMPv4 + icmp_payload_length, 1);
  
  	apsr_close_link_layer(apsr_ll);
        free(buf);
}

