#include "../include/apsrlib.h"

struct apsr_link_layer * apsr_open_link_layer (char *device) {
	#ifdef __OSTYPE_LINUX

	register struct apsr_link_layer *l;
	struct ifreq ifr;

	if((l=(struct apsr_link_layer *) xmalloc(sizeof(*l))) == NULL)
	{
	  #ifdef APSRLIB_DEBUG
	  ERROR("apsr_link_layer/sockpacket: xmalloc failed!");
	  #endif
	  
	  return(RET_NULL);
	}

	memset(l, 0, sizeof(*l));
	l->fd = socket(PF_INET, SOCK_PACKET, apsr_htons(APSR_ETH_P_ALL));

	if(l->fd <  0) {
	  #ifdef APSRLIB_DEBUG
	  ERROR("apsr_link_layer/sockpacket: socket failed!");
	  #endif
	
	  goto error;
	}

	memset(&ifr, 0, sizeof(ifr));
	strncpy(ifr.ifr_name, device, sizeof(ifr.ifr_name));

	if(ioctl(l->fd, SIOCGIFHWADDR, &ifr) < 0) {
	  #ifdef APSRLIB_DEBUG
	  ERROR("apsr_link_layer/sockpacket: ioctl failed!");
	  #endif

	  goto error;
	}

	switch (ifr.ifr_hwaddr.sa_family) {
	  case ARPHRD_ETHER:
	  case ARPHRD_METRICOM:
	  case ARPHRD_LOOPBACK:
	  	l->type = APSR_DLT_EN10MB;
		l->offset = 0xe;
		break;
	  case ARPHRD_EETHER:
		l->type = APSR_DLT_EN3MB;
	        break;	
          case ARPHRD_IEEE802_TR:
          case ARPHRD_IEEE802:
		l->type = APSR_DLT_IEEE802;
		l->offset = 0x2;
		break;
          case ARPHRD_TUNNEL:
          case ARPHRD_SIT:
          case ARPHRD_ADAPT:
  	  case ARPHRD_SLIP:
	  case ARPHRD_CSLIP:
	  case ARPHRD_SLIP6:
	  case ARPHRD_CSLIP6:
	  case ARPHRD_PPP:
          case ARPHRD_RAWHDLC:		
	  	l->type = APSR_DLT_RAW;
		break;
          case ARPHRD_IEEE80211:
		l->type = APSR_DLT_IEEE802_11;
		break;
          case ARPHRD_IEEE80211_PRISM:
                l->type = APSR_DLT_PRISM_HEADER;
                break;		
          case ARPHRD_AX25:
                l->type = APSR_DLT_AX25;
                break;
          case ARPHRD_PRONET:
                l->type = APSR_DLT_PRONET;
                break;
          case ARPHRD_CHAOS:
                l->type = APSR_DLT_CHAOS;
                break;
          case ARPHRD_ARCNET:
                l->type = APSR_DLT_ARCNET;
                break;
          case ARPHRD_FDDI:
                l->type = APSR_DLT_FDDI;
                l->offset = 0x3;
                break;
          case ARPHRD_HDLC:
                l->type = APSR_DLT_C_HDLC;
                break;
          case ARPHRD_LOCALTLK:
                l->type = APSR_DLT_LTALK;
                break;
 	  case ARPHRD_ATM:
		l->type = APSR_DLT_LINUX_SLL;
		break;
	  default:
	  	INFO("apsr_link_layer/sockpacket: unknown physical layer type!");
		goto error;
		break; /* Never reached */
	}

	return(l);

error:
	if(l->fd >= 0) close(l->fd);

	free(l);

	#endif
	
	return(RET_NULL);
}


int apsr_close_link_layer (struct apsr_link_layer *l) {
	#ifdef __OSTYPE_LINUX
	
	if(close(l->fd) == 0) return(RET_SUCCESS);
	else return(RET_FAILURE);

	return(RET_SUCCESS);

	#else

	return(RET_FAILURE);

	#endif
}


int apsr_write_link_layer (struct apsr_link_layer *l, char *device,
		unsigned char *buf, int len, int count)
{
	#ifdef __OSTYPE_LINUX

	int c=0;	/* Bytes sent through sendto.	*/
	int cnt;	/* Packet counter		*/
	struct sockaddr sa;

	memset(&sa, 0, sizeof(sa));
	strncpy(sa.sa_data, device, sizeof (sa.sa_data));

	/* Write the stuff down to the wire... */
	
	for(cnt=0; cnt < count; cnt++) {
		c += sendto(l->fd, buf, len, 0, (struct sockaddr *) &sa, sizeof(sa));

		if(c != len) {
		  #ifdef APSRLIB_DEBUG
		    INFO("apsr_write_link_layer: sendto failed!");
		  #endif
		}
	}

	return(c);

	#else

        return(RET_FAILURE);

        #endif
}

int apsr_read_link_layer (struct apsr_link_layer *l, int buf_size, int count,                             apsr_sniff_callback call_function) 
{
	#ifdef __OSTYPE_LINUX

	unsigned int i, len, sa_len;          
        struct sockaddr sa;
	unsigned char buf[65536];

        sa_len = sizeof(sa);
	
	if(count == 0) {
	 	while(1) {
	 		if((len = recvfrom(l->fd, (unsigned char *)buf, 
							buf_size, 0, 
				(struct sockaddr *) &sa, &sa_len)) == -1) {
                          #ifdef APSRLIB_DEBUG
                            INFO("apsr_read_link_layer: recv(loop) failed!");
                          #endif
                          apsr_close_link_layer(l);
                          exit(EXIT_FAILURE);
			}
			call_function(&sa, buf, len);	
	        }
        }
        else {
		for(i = 1; i <= count; i++) {
                        if((len = recvfrom(l->fd, (unsigned char *)buf, 
							buf_size, 0, 
				(struct sockaddr *) &sa, &sa_len)) == -1) {
          		  #ifdef APSRLIB_DEBUG
		            INFO("apsr_read_link_layer: recv(special count) failed!");
        		  #endif
			  apsr_close_link_layer(l);
			  exit(EXIT_FAILURE);
 			}
			call_function(&sa, buf, len);
	        }
	}

	return(RET_SUCCESS);
 
	#else
	
    	return(RET_FAILURE);

	#endif
}

